import importlib.metadata
from pathlib import Path

import colorama
import rich
import typer as typer
from PIL import Image
from rich.console import Console
import imageio as iio
from io import StringIO
from CliRenderer import render, Flags


def cli_main(source: Path = typer.Argument(..., help="The path to source image."),
             width: int = typer.Option(150, "--width", "-w", min=1,
                                       help="The width of the output in no. of characters."),
             height: int = typer.Option(50, "--height", "-h", min=1, help="The height of the output in no. of lines"),
             autosize: bool = typer.Option(False,
                                           help="Automatically sets the output size to fit the terminal. Overrides --width and --height. Only works on supported terminals."),

             output: Path = typer.Option(None, "--out", "-o",
                                         help="Saves the output to a file. (Written as bytes)"),
             debug: bool = typer.Option(False,
                                        help="Enables debug mode. This will save the intermediate images to the build folder."),
             bg_intensity: float = typer.Option(1.0, "--bg-intensity", "-bgi", min=0.0, max=1.0,
                                                help="The intensity of the background. The closer to 0 the darker the background.")
             ):
    """
    Renders the source image into the console as unicode art.
    """

    Flags.DEBUG = debug
    console = Console(file=StringIO(), record=True)

    if autosize:
        import shutil
        width, height = shutil.get_terminal_size()



    source_image = Image.fromarray(iio.v3.imread(source))
    text = render(source_image, (width, height), bg_intensity)

    console.print(text)
    string = console.export_text(styles=True)


    if output is not None:
        typer.echo(f"Saving output to {output}...")
        output.parent.mkdir(parents=True, exist_ok=True)
        with open(output, "wb") as f:
            f.write(string.encode("utf-8"))

    print(string)
    print(f"Saved output to {output}... Read it using `cat {output}`")
def main():
    rich.print(f"[bold white]CliRenderer installed version: {importlib.metadata.version('CliRenderer')}.[/bold white]")
    rich.print("[grey37]Version shown may not be accurate! (especially you are running from source!)[/grey37]\n")
    typer.run(cli_main)


if __name__ == "__main__":
    main()



import numpy as np
import numpy.typing as npt
from PIL import Image
import numba as nb
from CliRenderer.core import Flags

PixelsPerChar = (2, 4)
PixelsPerCharCount = PixelsPerChar[0] * PixelsPerChar[1]

braille6 = " ⠁⠂▘⠄⠅⠆⠇⠈▔⠊⠋⠌⠍⠎⠏⠐⠑⠒⠓⠔⠕⠖⠗▝⠙⠚▀⠜⠝⠞⠟⠠⠡⠢⠣⠤⠥⠦⠧⠨⠩⠪⠫⠬⠭⠮⠯⠰⠱⠲⠳⠴⠵⠶⠷⠸⠹⠺⠻⠼⠽⠾⠿"
braille8 = "⡀⡁⡂⡃▖⡅⡆▌⡈⡉⡊⡋⡌⡍⡎⡏⡐⡑⡒⡓⡔⡕⡖⡗⡘⡙⡚⡛▞⡝⡞▛⡠⡡⡢⡣⡤⡥⡦⡧⡨⡩⡪⡫⡬⡭⡮⡯⡰⡱⡲⡳⡴⡵⡶⡷⡸⡹⡺⡻⡼⡽⡾⡿⢀⢁⢂⢃⢄⢅⢆⢇⢈⢉⢊⢋⢌⢍⢎⢏⢐⢑⢒" \
           "⢓⢔⢕⢖⢗⢘⢙⢚⢛⢜⢝⢞⢟▗⢡⢢▚⢤⢥⢦⢧⢨⢩⢪⢫⢬⢭⢮⢯⢰⢱⢲⢳⢴⢵⢶⢷▐⢹⢺▜⢼⢽⢾⢿▂⣁⣂⣃⣄⣅⣆⣇⣈⣉⣊⣋⣌⣍⣎⣏⣐⣑⣒⣓⣔⣕⣖⣗⣘⣙⣚⣛⣜⣝⣞⣟⣠⣡⣢⣣▃⣥" \
           "⣦▙⣨⣩⣪⣫⣬⣭⣮⣯⣰⣱⣲⣳⣴⣵▅⣷⣸⣹⣺⣻▟⣽⣾█"
braille = braille6 + braille8


def pixels2Char(pixels: npt.NDArray):
    if len(pixels.shape) > 1 or len(pixels) != 8:
        raise Exception("Error, pixels provided must be in a 1D array of length 8!")

    # pixels = [1,1,1,1,1,1,1,1]
    converted = [pixels[7], pixels[3], pixels[6], pixels[5], pixels[4], pixels[2], pixels[1], pixels[0]]

    index = int("0b" + "".join("1" if x else "0" for x in converted), 2)

    return braille[index]

def generateColoredChars(image_: Image.Image) -> tuple[np.ndarray, np.ndarray]:
    """
    Splits an image up into PixelsPerChar sized chunks.
    Results in shape (rows,cols,8) last axis where 8 is the pixels in the chunk flattened into a 1D array
    :param image_:
    :return: Returns the image chunks (rows,cols,8) and a colors to use (rows,cols, (fore color, back color))
    """
    grayscale: npt.NDArray = np.asarray(image_.convert("L")).copy()
    sample: npt.NDArray = np.asarray(image_.convert("RGBA")).copy()

    rows = grayscale.shape[0] // PixelsPerChar[1]
    cols = grayscale.shape[1] // PixelsPerChar[0]
    # print(array.flags)
    if len(grayscale.shape) != 2:
        raise Exception("Error, image array is not grayscale!")

    data = np.zeros((rows, cols, PixelsPerCharCount), dtype=bool)
    colors = np.zeros((rows, cols, 2, sample.shape[2]), dtype=np.uint8)
    # print(sample.shape)
    # print(grayscale.shape)
    for y in range(rows):
        for x in range(cols):
            yCoords = (
                y * PixelsPerChar[1],
                (y + 1) * PixelsPerChar[1],
            )
            xCoords = (
                x * PixelsPerChar[0],
                (x + 1) * PixelsPerChar[0],
            )
            cell = grayscale[yCoords[0]:yCoords[1], xCoords[0]:xCoords[1]]

            min_index = int(np.argmin(cell))
            unraveled_min = np.unravel_index(min_index, PixelsPerChar[::-1])
            max_index = int(np.argmax(cell))
            unraveled_max = np.unravel_index(max_index, PixelsPerChar[::-1])
            # print(cell.shape)

            min_ = int(
                cell[unraveled_min[0], unraveled_min[1]])  # convert to python int to stop runtime overflow warnings
            max_ = int(cell[unraveled_max[0], unraveled_max[1]])

            mid = (min_ + max_) // 2
            bw = (cell > mid)

            data[y, x] = bw.flatten('F')

            colored_cell = sample[yCoords[0]:yCoords[1], xCoords[0]:xCoords[1]]

            colors[y, x, 0] = colored_cell[unraveled_max[0],unraveled_max[1]]
            colors[y, x, 1] = colored_cell[unraveled_min[0],unraveled_min[1]]

    if Flags.DEBUG:
        Image.fromarray(grayscale).save(f"./build/twotone.png")
        Image.fromarray(sample).save(f"./build/char-color.png")

    return data, colors



import os

from PIL import Image

from rich.text import Text

from CliRenderer.chartools import PixelsPerChar, generateColoredChars, pixels2Char
from CliRenderer.colorer import color_char
from CliRenderer.core import Flags


def render(source_: Image.Image, out_size=(170, 50), bg_intensity=1, skip_resize=False) -> Text:
    """
    Renders a single image into unicode text.

    :param source_: The source image
    :param out_size: The size of the final text in the form of (columns, lines).
    :return: returns a rich Text object to be printed to the console with rich
    :param skip_resize: If set to true, the image will not be resized. This is useful if the image is already the correct size.
    """

    FinalImageSize = (PixelsPerChar[0] * out_size[0], PixelsPerChar[1] * out_size[1])

    if not skip_resize:
        image = source_.resize(FinalImageSize)
    else:
        image = source_
    if image.size != FinalImageSize:
        raise ValueError("The image size does not match the output size. Cannot skip resize!")

    if Flags.DEBUG:
        os.makedirs("./build", exist_ok=True)
        image.save("./build/resized.png")


    data, charcolors = generateColoredChars(image)

    chars = []
    for row in data:
        char_row = ""
        for cell in row:
            char_row += pixels2Char(cell)
        chars.append(char_row)

    string = color_char( chars, charcolors, bg_intensity)
    return string
    # return ""


import init
from pathlib import Path
import time

from PIL import Image
from rich.console import Console

import CliRenderer


CliRenderer.Flags.DEBUG=False
Test_Count = 20

image = Image.open("../resources/pexels-pixabay-206359.jpg")
image2 = Image.open("../resources/pexels-photo-302769.jpeg")
console = Console()

timings = []

for i in range(Test_Count):
    start_time = time.time()
    txt = CliRenderer.render(image)
    resultA = (time.time()-start_time)*1000
    print(f"Test {i} A: {resultA} ms")

    start_time = time.time()
    txt2 = CliRenderer.render(image2)
    resultB = (time.time() - start_time) * 1000
    print(f"Test {i} B: {resultB} ms")

    timings.append(resultA)
    timings.append(resultB)

print(f"Average timings: {sum(timings)/len(timings)} ms")
print(f"Biggest: {max(timings)} ms")
print(f"Smallest: {min(timings)} ms")





"""
Contain functions related to coloring the text
"""

import numpy as np
from rich.color import Color
from rich.style import Style
from rich.text import Text

def color_char(chars: list[str], charcolors: np.ndarray, bg_intensity: float = 1):
    """
    Colors a string of text using the source image.

    :param bg_intensity: Intensity of the background color
    :param image: The source image
    :param chars: The string to color
    :param mask: The twotone mask
    :return:
    """

    string = Text()

    for y, row in enumerate(chars):
        for x, char in enumerate(row):
            rawfore_ = charcolors[y, x, 0]
            rawback_ = charcolors[y, x, 1]

            # Blend the back and fore colors in a 80:20 ratio so that the background color is effectively the 20% th percentile of cell colors
            # TLDR: The background color is the darkest color in the cell,
            # by blending it with the foreground color it looks nicer and is more representative of the cell color
            # The same goes for the foreground color (which is the brightest), but it is blended with the background color instead
            fore_ = rawfore_ * 0.8 + rawback_ * 0.2
            back_ = (rawfore_ * 0.2 + rawback_ * 0.8) * bg_intensity

            fore = Color.from_rgb(fore_[0], fore_[1], fore_[2])
            back = Color.from_rgb(back_[0], back_[1], back_[2])
            style = Style(color=fore, bold=True, bgcolor=back)
            string.append(char, style=style)
        string += "\r\n"

    return string
